@@ -19,6 +19,25 @@ package org.apache.maven.plugin.compiler;
  * under the License.
  */
 
+import java.io.File;
+import java.lang.reflect.InvocationTargetException;
+import java.lang.reflect.Method;
+import java.util.ArrayList;
+import java.util.Date;
+import java.util.HashSet;
+import java.util.LinkedHashSet;
+import java.util.List;
+import java.util.Map;
+import java.util.Set;
+
+import org.apache.maven.artifact.Artifact;
+import org.apache.maven.artifact.DefaultArtifact;
+import org.apache.maven.artifact.handler.ArtifactHandler;
+import org.apache.maven.artifact.handler.manager.ArtifactHandlerManager;
+import org.apache.maven.artifact.resolver.ArtifactResolutionRequest;
+import org.apache.maven.artifact.resolver.ArtifactResolutionResult;
+import org.apache.maven.artifact.resolver.ResolutionErrorHandler;
+import org.apache.maven.artifact.versioning.VersionRange;
 import org.apache.maven.execution.MavenSession;
 import org.apache.maven.plugin.AbstractMojo;
 import org.apache.maven.plugin.MojoExecution;
@@ -26,6 +45,7 @@ import org.apache.maven.plugin.MojoExecutionException;
 import org.apache.maven.plugins.annotations.Component;
 import org.apache.maven.plugins.annotations.Parameter;
 import org.apache.maven.project.MavenProject;
+import org.apache.maven.repository.RepositorySystem;
 import org.apache.maven.shared.incremental.IncrementalBuildHelper;
 import org.apache.maven.shared.incremental.IncrementalBuildHelperRequest;
 import org.apache.maven.shared.utils.ReaderFactory;
@@ -49,16 +69,6 @@ import org.codehaus.plexus.compiler.util.scan.mapping.SingleTargetSourceMapping;
 import org.codehaus.plexus.compiler.util.scan.mapping.SourceMapping;
 import org.codehaus.plexus.compiler.util.scan.mapping.SuffixMapping;
 
-import java.io.File;
-import java.lang.reflect.Method;
-import java.util.ArrayList;
-import java.util.Date;
-import java.util.HashSet;
-import java.util.LinkedHashMap;
-import java.util.List;
-import java.util.Map;
-import java.util.Set;
-
 /**
  * TODO: At least one step could be optimized, currently the plugin will do two
  * scans of all the source code if the compiler has to have the entire set of
@@ -67,13 +77,20 @@ import java.util.Set;
  *
  * @author others
  * @author <a href="mailto:trygvis@inamo.no">Trygve Laugst&oslash;l</a>
- * @version $Id: AbstractCompilerMojo.java 1629481 2014-10-05 09:27:03Z hboutemy $
+ * @version $Id: AbstractCompilerMojo.java 1796958 2017-05-30 23:23:58Z olamy $
  * @since 2.0
  */
 public abstract class AbstractCompilerMojo
     extends AbstractMojo
 {
 
+    static final String DEFAULT_SOURCE = "1.5";
+    
+    static final String DEFAULT_TARGET = "1.5";
+    
+    // Used to compare with older targets
+    static final String MODULE_INFO_TARGET = "1.9";
+    
     // ----------------------------------------------------------------------
     // Configurables
     // ----------------------------------------------------------------------
@@ -85,6 +102,14 @@ public abstract class AbstractCompilerMojo
      */
     @Parameter( property = "maven.compiler.failOnError", defaultValue = "true" )
     private boolean failOnError = true;
+    
+    /**
+     * Indicates whether the build will continue even if there are compilation warnings.
+     *
+     * @since 3.6
+     */
+    @Parameter( property = "maven.compiler.failOnWarning", defaultValue = "false" )
+    private boolean failOnWarning;  
 
     /**
      * Set to <code>true</code> to include debugging information in the compiled class files.
@@ -92,6 +117,13 @@ public abstract class AbstractCompilerMojo
     @Parameter( property = "maven.compiler.debug", defaultValue = "true" )
     private boolean debug = true;
 
+    /**
+     * Set to <code>true</code> to generate metadata for reflection on method parameters.
+     * @since 3.6.2
+     */
+    @Parameter( property = "maven.compiler.parameters", defaultValue = "false" )
+    private boolean parameters;
+
     /**
      * Set to <code>true</code> to show messages about what the compiler is doing.
      */
@@ -119,15 +151,23 @@ public abstract class AbstractCompilerMojo
     /**
      * The -source argument for the Java compiler.
      */
-    @Parameter( property = "maven.compiler.source", defaultValue = "1.5" )
+    @Parameter( property = "maven.compiler.source", defaultValue = DEFAULT_SOURCE )
     protected String source;
 
     /**
      * The -target argument for the Java compiler.
      */
-    @Parameter( property = "maven.compiler.target", defaultValue = "1.5" )
+    @Parameter( property = "maven.compiler.target", defaultValue = DEFAULT_TARGET )
     protected String target;
 
+    /**
+     * The -release argument for the Java compiler, supported since Java9
+     * 
+     * @since 3.6
+     */
+    @Parameter( property = "maven.compiler.release" )
+    protected String release;
+    
     /**
      * The -encoding argument for the Java compiler.
      *
@@ -214,17 +254,46 @@ public abstract class AbstractCompilerMojo
     @Parameter
     private String[] annotationProcessors;
 
+    /**
+     * <p>
+     * Classpath elements to supply as annotation processor path. If specified, the compiler will detect annotation
+     * processors only in those classpath elements. If omitted, the default classpath is used to detect annotation
+     * processors. The detection itself depends on the configuration of {@code annotationProcessors}.
+     * </p>
+     * <p>
+     * Each classpath element is specified using their Maven coordinates (groupId, artifactId, version, classifier,
+     * type). Transitive dependencies are added automatically. Example:
+     * </p>
+     *
+     * <pre>
+     * &lt;configuration&gt;
+     *   &lt;annotationProcessorPaths&gt;
+     *     &lt;path&gt;
+     *       &lt;groupId&gt;org.sample&lt;/groupId&gt;
+     *       &lt;artifactId&gt;sample-annotation-processor&lt;/artifactId&gt;
+     *       &lt;version&gt;1.2.3&lt;/version&gt;
+     *     &lt;/path&gt;
+     *     &lt;!-- ... more ... --&gt;
+     *   &lt;/annotationProcessorPaths&gt;
+     * &lt;/configuration&gt;
+     * </pre>
+     *
+     * @since 3.5
+     */
+    @Parameter
+    private List<DependencyCoordinate> annotationProcessorPaths;
+
     /**
      * <p>
      * Sets the arguments to be passed to the compiler (prepending a dash) if {@link #fork} is set to <code>true</code>.
      * </p>
      * <p>
-     * This is because the list of valid arguments passed to a Java compiler
-     * varies based on the compiler version.
+     * This is because the list of valid arguments passed to a Java compiler varies based on the compiler version.
      * </p>
      * <p>
      * To pass <code>-Xmaxerrs 1000 -Xlint -Xlint:-path -Averbose=true</code> you should include the following:
      * </p>
+     *
      * <pre>
      * &lt;compilerArguments&gt;
      *   &lt;Xmaxerrs&gt;1000&lt;/Xmaxerrs&gt;
@@ -235,7 +304,7 @@ public abstract class AbstractCompilerMojo
      * </pre>
      *
      * @since 2.0.1
-     * @deprecated use {@link #compilerArgs} instead. 
+     * @deprecated use {@link #compilerArgs} instead.
      */
     @Parameter
     @Deprecated
@@ -257,7 +326,7 @@ public abstract class AbstractCompilerMojo
      */
     @Parameter
     protected List<String> compilerArgs;
-    
+
     /**
      * <p>
      * Sets the unformatted single argument string to be passed to the compiler if {@link #fork} is set to
@@ -297,6 +366,18 @@ public abstract class AbstractCompilerMojo
     @Component
     private ToolchainManager toolchainManager;
 
+    /**
+     * <p>
+     * Specify the requirements for this jdk toolchain.
+     * This overrules the toolchain selected by the maven-toolchain-plugin.
+     * </p>
+     * <strong>note:</strong> requires at least Maven 3.3.1
+     * 
+     * @since 3.6
+     */
+    @Parameter
+    private Map<String, String> jdkToolchain;
+
     // ----------------------------------------------------------------------
     // Read-only parameters
     // ----------------------------------------------------------------------
@@ -366,7 +447,7 @@ public abstract class AbstractCompilerMojo
     /**
      * @since 3.0 needed for storing the status for the incremental build support.
      */
-    @Parameter( property = "mojoExecution", readonly = true, required = true )
+    @Parameter( defaultValue = "${mojoExecution}", readonly = true, required = true )
     private MojoExecution mojoExecution;
 
     /**
@@ -385,13 +466,35 @@ public abstract class AbstractCompilerMojo
     @Parameter( defaultValue = "true", property = "maven.compiler.useIncrementalCompilation" )
     private boolean useIncrementalCompilation = true;
 
+    /**
+     * Resolves the artifacts needed.
+     */
+    @Component
+    private RepositorySystem repositorySystem;
+
+    /**
+     * Artifact handler manager.
+     */
+    @Component
+    private ArtifactHandlerManager artifactHandlerManager;
+
+    /**
+     * Throws an exception on artifact resolution errors.
+     */
+    @Component
+    private ResolutionErrorHandler resolutionErrorHandler;
+
     protected abstract SourceInclusionScanner getSourceInclusionScanner( int staleMillis );
 
     protected abstract SourceInclusionScanner getSourceInclusionScanner( String inputFileEnding );
 
     protected abstract List<String> getClasspathElements();
 
+    protected abstract List<String> getModulepathElements();
+
     protected abstract List<String> getCompileSourceRoots();
+    
+    protected abstract void preparePaths( Set<File> sourceFiles );
 
     protected abstract File getOutputDirectory();
 
@@ -399,12 +502,20 @@ public abstract class AbstractCompilerMojo
 
     protected abstract String getTarget();
 
+    protected abstract String getRelease();
+
     protected abstract String getCompilerArgument();
 
     protected abstract Map<String, String> getCompilerArguments();
 
     protected abstract File getGeneratedSourcesDirectory();
 
+    protected final MavenProject getProject()
+    {
+        return project;
+    }
+
+    @Override
     public void execute()
         throws MojoExecutionException, CompilationFailureException
     {
@@ -432,7 +543,7 @@ public abstract class AbstractCompilerMojo
         Toolchain tc = getToolchain();
         if ( tc != null )
         {
-            getLog().info( "Toolchain in compiler-plugin: " + tc );
+            getLog().info( "Toolchain in maven-compiler-plugin: " + tc );
             if ( executable != null )
             {
                 getLog().warn( "Toolchains are ignored, 'executable' parameter is set to " + executable );
@@ -457,13 +568,6 @@ public abstract class AbstractCompilerMojo
             return;
         }
 
-        if ( getLog().isDebugEnabled() )
-        {
-            getLog().debug( "Source directories: " + compileSourceRoots.toString().replace( ',', '\n' ) );
-            getLog().debug( "Classpath: " + getClasspathElements().toString().replace( ',', '\n' ) );
-            getLog().debug( "Output directory: " + getOutputDirectory() );
-        }
-
         // ----------------------------------------------------------------------
         // Create the compiler configuration
         // ----------------------------------------------------------------------
@@ -472,8 +576,6 @@ public abstract class AbstractCompilerMojo
 
         compilerConfiguration.setOutputLocation( getOutputDirectory().getAbsolutePath() );
 
-        compilerConfiguration.setClasspathEntries( getClasspathElements() );
-
         compilerConfiguration.setOptimize( optimize );
 
         compilerConfiguration.setDebug( debug );
@@ -493,20 +595,27 @@ public abstract class AbstractCompilerMojo
             compilerConfiguration.setDebugLevel( debuglevel );
         }
 
+        compilerConfiguration.setParameters( parameters );
+
         compilerConfiguration.setVerbose( verbose );
 
         compilerConfiguration.setShowWarnings( showWarnings );
 
+        compilerConfiguration.setFailOnWarning( failOnWarning );
+
         compilerConfiguration.setShowDeprecation( showDeprecation );
 
         compilerConfiguration.setSourceVersion( getSource() );
 
         compilerConfiguration.setTargetVersion( getTarget() );
+        
+        compilerConfiguration.setReleaseVersion( getRelease() );
 
         compilerConfiguration.setProc( proc );
 
         File generatedSourcesDirectory = getGeneratedSourcesDirectory();
-        compilerConfiguration.setGeneratedSourcesDirectory( generatedSourcesDirectory );
+        compilerConfiguration.setGeneratedSourcesDirectory( generatedSourcesDirectory != null
+                        ? generatedSourcesDirectory.getAbsoluteFile() : null );
 
         if ( generatedSourcesDirectory != null )
         {
@@ -543,51 +652,10 @@ public abstract class AbstractCompilerMojo
 
         compilerConfiguration.setAnnotationProcessors( annotationProcessors );
 
+        compilerConfiguration.setProcessorPathEntries( resolveProcessorPathEntries() );
+
         compilerConfiguration.setSourceEncoding( encoding );
 
-        Map<String, String> effectiveCompilerArguments = getCompilerArguments();
-
-        String effectiveCompilerArgument = getCompilerArgument();
-
-        if ( ( effectiveCompilerArguments != null ) || ( effectiveCompilerArgument != null )
-                        || ( compilerArgs != null ) )
-        {
-            LinkedHashMap<String, String> cplrArgsCopy = new LinkedHashMap<String, String>();
-            if ( effectiveCompilerArguments != null )
-            {
-                for ( Map.Entry<String, String> me : effectiveCompilerArguments.entrySet() )
-                {
-                    String key = me.getKey();
-                    String value = me.getValue();
-                    if ( !key.startsWith( "-" ) )
-                    {
-                        key = "-" + key;
-                    }
-
-                    if ( key.startsWith( "-A" ) && StringUtils.isNotEmpty( value ) )
-                    {
-                        cplrArgsCopy.put( key + "=" + value, null );
-                    }
-                    else
-                    {
-                        cplrArgsCopy.put( key, value );
-                    }
-                }
-            }
-            if ( !StringUtils.isEmpty( effectiveCompilerArgument ) )
-            {
-                cplrArgsCopy.put( effectiveCompilerArgument, null );
-            }
-            if ( compilerArgs != null )
-            {
-                for ( String arg : compilerArgs )
-                {
-                    cplrArgsCopy.put( arg, null );
-                }
-            }
-            compilerConfiguration.setCustomCompilerArguments( cplrArgsCopy );
-        }
-
         compilerConfiguration.setFork( fork );
 
         if ( fork )
@@ -679,13 +747,17 @@ public abstract class AbstractCompilerMojo
                 canUpdateTarget = compiler.canUpdateTarget( compilerConfiguration );
 
                 sources = getCompileSources( compiler, compilerConfiguration );
+                
+                preparePaths( sources );
 
                 incrementalBuildHelperRequest = new IncrementalBuildHelperRequest().inputFiles( sources );
 
+                // CHECKSTYLE_OFF: LineLength
                 if ( ( compiler.getCompilerOutputStyle().equals( CompilerOutputStyle.ONE_OUTPUT_FILE_FOR_ALL_INPUT_FILES ) && !canUpdateTarget )
                     || isDependencyChanged()
                     || isSourceChanged( compilerConfiguration, compiler )
                     || incrementalBuildHelper.inputFileTreeChanged( incrementalBuildHelperRequest ) )
+                    // CHECKSTYLE_ON: LineLength
                 {
                     getLog().info( "Changes detected - recompiling the module!" );
 
@@ -730,6 +802,8 @@ public abstract class AbstractCompilerMojo
                 {
                     compilerConfiguration.setSourceFiles( staleSources );
                 }
+                
+                preparePaths( compilerConfiguration.getSourceFiles() );
             }
             catch ( CompilerException e )
             {
@@ -743,10 +817,56 @@ public abstract class AbstractCompilerMojo
                 return;
             }
         }
+        
+        // Dividing pathElements of classPath and modulePath is based on sourceFiles
+        compilerConfiguration.setClasspathEntries( getClasspathElements() );
+
+        compilerConfiguration.setModulepathEntries( getModulepathElements() );
+        
+        Map<String, String> effectiveCompilerArguments = getCompilerArguments();
+
+        String effectiveCompilerArgument = getCompilerArgument();
+
+        if ( ( effectiveCompilerArguments != null ) || ( effectiveCompilerArgument != null )
+                        || ( compilerArgs != null ) )
+        {
+            if ( effectiveCompilerArguments != null )
+            {
+                for ( Map.Entry<String, String> me : effectiveCompilerArguments.entrySet() )
+                {
+                    String key = me.getKey();
+                    String value = me.getValue();
+                    if ( !key.startsWith( "-" ) )
+                    {
+                        key = "-" + key;
+                    }
+
+                    if ( key.startsWith( "-A" ) && StringUtils.isNotEmpty( value ) )
+                    {
+                        compilerConfiguration.addCompilerCustomArgument( key + "=" + value, null );
+                    }
+                    else
+                    {
+                        compilerConfiguration.addCompilerCustomArgument( key, value );
+                    }
+                }
+            }
+            if ( !StringUtils.isEmpty( effectiveCompilerArgument ) )
+            {
+                compilerConfiguration.addCompilerCustomArgument( effectiveCompilerArgument, null );
+            }
+            if ( compilerArgs != null )
+            {
+                for ( String arg : compilerArgs )
+                {
+                    compilerConfiguration.addCompilerCustomArgument( arg, null );
+                }
+            }
+        }
+
         // ----------------------------------------------------------------------
         // Dump configuration
         // ----------------------------------------------------------------------
-
         if ( getLog().isDebugEnabled() )
         {
             getLog().debug( "Classpath:" );
@@ -756,6 +876,15 @@ public abstract class AbstractCompilerMojo
                 getLog().debug( " " + s );
             }
 
+            if ( !getModulepathElements().isEmpty() )
+            {
+                getLog().debug( "Modulepath:" );
+                for ( String s : getModulepathElements() )
+                {
+                    getLog().debug( " " + s );
+                }
+            }
+
             getLog().debug( "Source roots:" );
 
             for ( String root : getCompileSourceRoots() )
@@ -775,7 +904,7 @@ public abstract class AbstractCompilerMojo
                 }
 
                 String[] cl = compiler.createCommandLine( compilerConfiguration );
-                if ( cl != null && cl.length > 0 )
+                if ( getLog().isDebugEnabled() && cl != null && cl.length > 0 )
                 {
                     StringBuilder sb = new StringBuilder();
                     sb.append( cl[0] );
@@ -991,7 +1120,8 @@ public abstract class AbstractCompilerMojo
         {
             File rootFile = new File( sourceRoot );
 
-            if ( !rootFile.isDirectory() )
+            if ( !rootFile.isDirectory()
+                || rootFile.getAbsoluteFile().equals( compilerConfiguration.getGeneratedSourcesDirectory() ) )
             {
                 continue;
             }
@@ -1056,20 +1186,25 @@ public abstract class AbstractCompilerMojo
 
     protected Date getBuildStartTime()
     {
+        Date buildStartTime = null;
         try
         {
             Method getRequestMethod = session.getClass().getMethod( "getRequest" );
             Object mavenExecutionRequest = getRequestMethod.invoke( session );
             Method getStartTimeMethod = mavenExecutionRequest.getClass().getMethod( "getStartTime" );
-            Date buildStartTime = (Date) getStartTimeMethod.invoke( mavenExecutionRequest );
-            return buildStartTime;
+            buildStartTime = (Date) getStartTimeMethod.invoke( mavenExecutionRequest );
         }
         catch ( Exception e )
         {
             getLog().debug( "unable to get start time for the current build: " + e.getMessage() );
         }
 
-        return new Date();
+        if ( buildStartTime == null )
+        {
+            return new Date();
+        }
+
+        return buildStartTime;
     }
 
 
@@ -1095,10 +1230,53 @@ public abstract class AbstractCompilerMojo
     private Toolchain getToolchain()
     {
         Toolchain tc = null;
-        if ( toolchainManager != null )
+        
+        if ( jdkToolchain != null )
+        {
+            // Maven 3.3.1 has plugin execution scoped Toolchain Support
+            try
+            {
+                Method getToolchainsMethod =
+                    toolchainManager.getClass().getMethod( "getToolchains", MavenSession.class, String.class,
+                                                           Map.class );
+
+                @SuppressWarnings( "unchecked" )
+                List<Toolchain> tcs =
+                    (List<Toolchain>) getToolchainsMethod.invoke( toolchainManager, session, "jdk",
+                                                                  jdkToolchain );
+
+                if ( tcs != null && tcs.size() > 0 )
+                {
+                    tc = tcs.get( 0 );
+                }
+            }
+            catch ( NoSuchMethodException e )
+            {
+                // ignore
+            }
+            catch ( SecurityException e )
+            {
+                // ignore
+            }
+            catch ( IllegalAccessException e )
+            {
+                // ignore
+            }
+            catch ( IllegalArgumentException e )
+            {
+                // ignore
+            }
+            catch ( InvocationTargetException e )
+            {
+                // ignore
+            }
+        }
+        
+        if ( tc == null )
         {
             tc = toolchainManager.getToolchainFromBuildContext( "jdk", session );
         }
+        
         return tc;
     }
 
@@ -1204,10 +1382,9 @@ public abstract class AbstractCompilerMojo
     }
 
     /**
-     * We just compare the timestamps of all local dependency files (inter-module dependency classpath)
-     * and the own generated classes
-     * and if we got a file which is >= the buid-started timestamp, then we catched a file which got
-     * changed during this build.
+     * We just compare the timestamps of all local dependency files (inter-module dependency classpath) and the own
+     * generated classes and if we got a file which is &gt;= the buid-started timestamp, then we catched a file which
+     * got changed during this build.
      *
      * @return <code>true</code> if at least one single dependency has changed.
      */
@@ -1228,11 +1405,15 @@ public abstract class AbstractCompilerMojo
 
         Date buildStartTime = getBuildStartTime();
 
-        for ( String classPathElement : getClasspathElements() )
+        List<String> pathElements = new ArrayList<String>();
+        pathElements.addAll( getClasspathElements() );
+        pathElements.addAll( getModulepathElements() );
+        
+        for ( String pathElement : pathElements )
         {
             // ProjectArtifacts are artifacts which are available in the local project
             // that's the only ones we are interested in now.
-            File artifactPath = new File( classPathElement );
+            File artifactPath = new File( pathElement );
             if ( artifactPath.isDirectory() )
             {
                 if ( hasNewFile( artifactPath, buildStartTime ) )
@@ -1278,4 +1459,60 @@ public abstract class AbstractCompilerMojo
         return false;
     }
 
+    private List<String> resolveProcessorPathEntries()
+        throws MojoExecutionException
+    {
+        if ( annotationProcessorPaths == null || annotationProcessorPaths.isEmpty() )
+        {
+            return null;
+        }
+
+        try
+        {
+            Set<Artifact> requiredArtifacts = new LinkedHashSet<Artifact>();
+
+            for ( DependencyCoordinate coord : annotationProcessorPaths )
+            {
+                ArtifactHandler handler = artifactHandlerManager.getArtifactHandler( coord.getType() );
+
+                Artifact artifact = new DefaultArtifact(
+                     coord.getGroupId(),
+                     coord.getArtifactId(),
+                     VersionRange.createFromVersionSpec( coord.getVersion() ),
+                     Artifact.SCOPE_RUNTIME,
+                     coord.getType(),
+                     coord.getClassifier(),
+                     handler,
+                     false );
+
+                requiredArtifacts.add( artifact );
+            }
+
+            ArtifactResolutionRequest request = new ArtifactResolutionRequest()
+                            .setArtifact( requiredArtifacts.iterator().next() )
+                            .setResolveRoot( true )
+                            .setResolveTransitively( true )
+                            .setArtifactDependencies( requiredArtifacts )
+                            .setLocalRepository( session.getLocalRepository() )
+                            .setRemoteRepositories( project.getRemoteArtifactRepositories() );
+
+            ArtifactResolutionResult resolutionResult = repositorySystem.resolve( request );
+
+            resolutionErrorHandler.throwErrors( request, resolutionResult );
+
+            List<String> elements = new ArrayList<String>( resolutionResult.getArtifacts().size() );
+
+            for ( Object resolved : resolutionResult.getArtifacts() )
+            {
+                elements.add( ( (Artifact) resolved ).getFile().getAbsolutePath() );
+            }
+
+            return elements;
+        }
+        catch ( Exception e )
+        {
+            throw new MojoExecutionException( "Resolution of annotationProcessorPath dependencies failed: "
+                + e.getLocalizedMessage(), e );
+        }
+    }
 }
